/* Name: main.c
 * Project: EasyLogger
 * Author: Christian Starkjohann
 * Creation Date: 2006-04-23
 * Tabsize: 4
 * Copyright: (c) 2006 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: Proprietary, free under certain conditions. See Documentation.
 * This Revision: $Id$
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "usbdrv.h"
#include "softuart.h"
#include "oddebug.h"

/* interface with usb_control_msg for CLI */
#define CLICMD_ECHO   0
#define CLICMD_GETOSC 1
#define CLICMD_SETOSC 2
#define CLICMD_HITKEY 3
#define CLICMD_GETCLID 4
#define CLICMD_CLRCLID 5

#define STATE_WAIT 0
#define STATE_SEND_KEY 1
#define STATE_RELEASE_KEY 2

#define UTIL_BIN4(x)        (uchar)((0##x & 01000)/64 + (0##x & 0100)/16 + (0##x & 010)/4 + (0##x & 1))
#define UTIL_BIN8(hi, lo)   (uchar)(UTIL_BIN4(hi) * 16 + UTIL_BIN4(lo))

#ifndef NULL
#define NULL    ((void *)0)
#endif

/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */
static uchar    defOSCCAL, state = STATE_WAIT;
static  char    cliddata[30];       /* 8 timestamp + max 20 calling line identity + 1 end of string character + 1 unused */


/* ------------------------------------------------------------------------- */

const PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */
#define MOD_CONTROL_LEFT    (1<<0)
#define MOD_SHIFT_LEFT      (1<<1)
#define MOD_ALT_LEFT        (1<<2)
#define MOD_GUI_LEFT        (1<<3)
#define MOD_CONTROL_RIGHT   (1<<4)
#define MOD_SHIFT_RIGHT     (1<<5)
#define MOD_ALT_RIGHT       (1<<6)
#define MOD_GUI_RIGHT       (1<<7)

#define KEY_OKAY      0x4B

/* ------------------------------------------------------------------------- */

static void buildReport(void)
{
    reportBuffer[0] = 0;    /* no modifiers */
    if(state == STATE_SEND_KEY)
        reportBuffer[1] = KEY_OKAY;
    else
        reportBuffer[1] = 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* ------------------------------------------------------------------------- */

uchar	usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;
static uchar    replyBuf[2];

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        switch(rq->bRequest) {
        case USBRQ_HID_GET_REPORT: // send "no keys pressed" if asked here
            // wValue: ReportType (highbyte), ReportID (lowbyte)
            usbMsgPtr = reportBuffer;
            reportBuffer[0] = 0;
            reportBuffer[1] = 0;
            return sizeof(reportBuffer);
        case USBRQ_HID_GET_IDLE: // send idle rate to PC as required by spec
            usbMsgPtr = &idleRate;
            return 1;
        case USBRQ_HID_SET_IDLE: // save idle rate as required by spec
            idleRate = rq->wValue.bytes[1];
            return 0;
        }
    }else{
        switch(rq->bRequest) {
        case CLICMD_ECHO:  /* result = 2 bytes */
            usbMsgPtr = replyBuf;
            replyBuf[0] = rq->wValue.bytes[0];
            replyBuf[1] = rq->wValue.bytes[1];
            return 2;
        case CLICMD_GETOSC:  /* result = 2 bytes */
            usbMsgPtr = replyBuf;
            replyBuf[0] = defOSCCAL;
            replyBuf[1] = OSCCAL;
            return 2;
        case CLICMD_SETOSC:  /* no response expected */
            OSCCAL = rq->wIndex.bytes[0];
            return 0;
        case CLICMD_HITKEY:  /* no response expected */
            state = STATE_SEND_KEY;
            return 0;
        case CLICMD_GETCLID:  /* result = sizeof(cliddata) */
            usbMsgPtr = cliddata;
            return sizeof(cliddata);
        case CLICMD_CLRCLID:  /* no response expected */
            cliddata[0] = 0;
            return 0;
        }
    }
    return 0;
}

void clid(void)
{
static uchar    data, message_transfer, message_len, byte_num, checksum, param_len, param_index, clid_len;

    data = softuart_getchar();

    /* message_transfer
       = 0 if message has not started
       = 1 if correct message type (80h) received
       = 2 if message length received and ready for parameter type for date and time (01h)
       = 3 if parameter type for date and time (01h) received
       = 4 if parameter length for date and time received
       = 5 if date and time received and ready for parameter type or checksum
       = 6 if parameter type for calling line identity (02h) received
       = 7 if parameter length for calling line identity received
       = 8 if parameter type for reason for absence (04h/08h) received
       = 9 if parameter length for reason for absence received */

    switch(message_transfer){
    case 0: // message has not started
        byte_num = 0;
        message_len = 2;
        clid_len = 0;
        checksum = 0;
        if(data == 0x80)
            message_transfer++;
        break;
    case 1: // correct message type received
        message_len = data + 2;
        if(message_len>=15)
            message_transfer++;
        else
            message_transfer=0;
        break;
    case 2: // message length received and ready for parameter type for date and time (01h)
        if(data == 0x01)
            message_transfer++;
        else
            message_transfer=0;
        break;
    case 3: // parameter type for date and time (01h) received
        param_len = data;
        param_index = 0;
        if(param_len == 8)
            message_transfer++;
        else
            message_transfer=0;
        break;
    case 4: // parameter length for date and time received
        if (data>='0' && data<='9')
            cliddata[param_index] = data;
        else
            message_transfer=0;
        if(++param_index == param_len)
            message_transfer = 5;
        break;
    case 5: // date and time received and ready for parameter type or checksum
        if(byte_num != message_len)
            switch(data){
            case 2:
                message_transfer = 6;
                break;
            case 4:
            case 8:
                message_transfer = 8;
                break;
            default:
                message_transfer = 0;
            }
        break;
    case 6: // parameter type for calling line identity (02h) received
        param_len = data;
        clid_len = data;
        param_index = 0;
        if(param_len == 0)
            message_transfer = 5;
        else
            if(param_len <= 20)
                message_transfer++;
            else
                message_transfer = 0;
	break;
    case 7: // parameter length for calling line identity received
        if (data>='0' && data<='9')
            cliddata[param_index + 8] = data;
        else
            message_transfer=0;
        if(++param_index == param_len){
            cliddata[param_len + 8] = 0;
            message_transfer = 5;
        }
        break;
    case 8: // parameter type for reason for absence (04h/08h) received
        if(data == 0x01)
            message_transfer++;
	else
            message_transfer = 0;
	break;
    case 9: // parameter length for reason for absence received
        if(data == 'O' || data == 'P'){
            message_transfer = 5;
            if(clid_len == 0){
                cliddata[8] = data;
                cliddata[9] = 0;
            }
        }else
            message_transfer = 0;
        break;
    }
    if(message_transfer != 0){
        checksum = checksum + data;
        if(byte_num == message_len){
            if(checksum == 0)
                // checksum correct: message received
                state = STATE_SEND_KEY;
            message_transfer = 0; // ready for new message
        }else
            byte_num++;
    }
}


/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET.
 * More info: /home/sil/avr/vusb/projects/tinycalibration/readme.txt
 */
static void calibrateOscillator(void)
{
uchar       step = 64;
uchar       trialValue, optimumValue;
int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

    if(OSCCAL < 116)         /* pre-programmed calibration value OSCCAL determines range to search */
        trialValue = 0;
    else if(OSCCAL < 128)    /* ranges are overlapping and this value is very high for the low range */
	trialValue = 128;    /*  so search high range */
    else if(OSCCAL < 140)    /* ranges are overlapping and this value is very low for the high range */
	trialValue = 0;      /*  so search low range */
    else
        trialValue = 128;
    /* do a binary search: */
    do{
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    /* proportional to current real frequency */
        if(x < targetValue)             /* frequency still too low */
            trialValue += step;
        step >>= 1;
    }while(step > 0);
    /* We have a precision of +/- 1 for optimum OSCCAL here */
    /* now do a neighborhood search for optimum value */
    optimumValue = trialValue;
    optimumDev = x; /* this is certainly far away from optimum */
    for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
        x = usbMeasureFrameLength() - targetValue;
        if(x < 0)
            x = -x;
        if(x < optimumDev){
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}

void    usbEventResetReady(void)
{
    /* Disable interrupts during oscillator calibration since
     * usbMeasureFrameLength() counts CPU cycles.
     */
    cli();
    calibrateOscillator();
    sei();
}

/* ------------------------------------------------------------------------- */
/* --------------------------------- main ---------------------------------- */
/* ------------------------------------------------------------------------- */

int main(void)
{
uchar   i;

    defOSCCAL=OSCCAL;

    /* calibration value OSCCAL is fine tuned after USB reset. Refer to Oscillator Calibration above */

    odDebugInit();
    usbDeviceDisconnect();
    for(i=0;i<20;i++){  /* 300 ms disconnect */
        _delay_ms(15);
    }
    usbDeviceConnect();
    usbInit();
    softuart_init();
    sei();
    for(;;){    /* main event loop */
        if(softuart_kbhit())
            /* process input character received by softuart */
            clid();
        usbPoll();
        if(usbInterruptIsReady() && state != STATE_WAIT){
            switch(state) {
            case STATE_SEND_KEY:
                buildReport();
                state = STATE_RELEASE_KEY; // release next
                break;
            case STATE_RELEASE_KEY:
                buildReport();
                state = STATE_WAIT; // go back to waiting
                break;
            default:
                state = STATE_WAIT; // should not happen
            }
            // start sending
            usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
        }
    }
    return 0;
}
